from math import floor

from mediakit.info import name, version
from mediakit.globals import global_config
from mediakit.utils.format import limit_text_length, parse_int
from mediakit.media.download import DownloadStatusCodes
from mediakit.cli.colors import Colors, colored
from mediakit.cli.screen import Content, ContentCategories, screen
from mediakit.cli.download.download_cli_store import DownloadCLIStore

BYTES_PER_MEGABYTES = 1024 * 1024

LOADING_DOTS = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
ROTATING_LINES = ["|", "/", "-", "\\"]


class DownloadCLIFormatter:
    def __init__(self, store: DownloadCLIStore):
        self.store = store

    def format_header(self):
        header = colored(f"{name.lower()} v{version}\n", style=Colors.style.BRIGHT)
        return header

    def format_batch_file_info_header(self, batch_filename: str):
        return (
            "\nReading URLs from "
            + colored(batch_filename, fore=Colors.fore.CYAN, style=Colors.style.BRIGHT)
            + "...\n"
        )

    def format_batch_file_info_body(self, valid_video_urls_in_batch_file):
        number_of_valid_urls = len(valid_video_urls_in_batch_file)

        return (
            "Found "
            + colored(
                f"{number_of_valid_urls} ",
                fore=Colors.fore.CYAN,
                style=Colors.style.BRIGHT,
            )
            + (
                colored("video URLs", fore=Colors.fore.CYAN)
                if number_of_valid_urls > 1
                else colored("video URL", fore=Colors.fore.CYAN)
            )
            + ". Preparing to download...\n"
        )

    def format_download_summary(self, video_index: int):
        skipped_formats_warning = self.format_skipped_formats_warning(video_index)
        fallback_replacement_summary = self.format_fallback_replacement_summary(
            video_index
        )

        total_download_size_label = self.format_total_download_size_label(video_index)

        download_summary_components = [
            self.format_video_heading(video_index) + "\n",
            skipped_formats_warning,
            fallback_replacement_summary,
            total_download_size_label,
        ]

        return "".join(
            filter(lambda component: len(component) > 0, download_summary_components)
        )

    def format_video_heading(self, video_index: int):
        formatted_video_length = self.format_video_length(video_index)

        remaining_number_of_characters = (
            min(screen.get_console_width(), self.store.MAX_SCREEN_WIDTH)
            - len(formatted_video_length)
            - 4
        )

        formatted_video_title = self.format_video_title(
            video_index, remaining_number_of_characters
        )

        spaces_between_heading_components = " " * (
            min(screen.get_console_width(), self.store.MAX_SCREEN_WIDTH)
            - len(formatted_video_title)
            - len(formatted_video_length)
            - 3
        )

        video_heading = (
            "\n"
            + colored(
                formatted_video_title, fore=Colors.fore.CYAN, style=Colors.style.BRIGHT
            )
            + spaces_between_heading_components
            + colored(f"({formatted_video_length})\n", style=Colors.style.BRIGHT)
        )

        return video_heading

    def format_video_length(self, video_index: int):
        seconds_section = self.store.videos[video_index].length % 60
        remaining_minutes = self.store.videos[video_index].length // 60
        minutes_section = remaining_minutes % 60
        hours_section = remaining_minutes // 60

        formatted_video_length = (
            (f"{hours_section}:" if hours_section > 0 else "")
            + f"{minutes_section:0>2}:"
            + f"{seconds_section:0>2}"
        )

        return formatted_video_length

    def format_video_title(
        self, video_index: int, limit_of_characters: int = None
    ) -> str:
        return limit_text_length(
            self.store.videos[video_index].title,
            self.store.MAX_SHORT_VIDEO_TITLE_LENGTH
            if limit_of_characters is None
            else limit_of_characters,
        )

    def format_download_confirmation_prompt_message(self):
        download_confirmation_prompt = "\nConfirm download? " + colored(
            "(Y/n) ",
            fore=Colors.fore.CYAN,
        )

        return download_confirmation_prompt

    def format_current_download_progress(self, video_index: int):
        media_resource = self.store.downloading_media_resources[video_index]

        heading = self.format_download_progress_heading(video_index)
        current_download_progress = heading + "\n"

        if media_resource.download_status in DownloadStatusCodes.CONVERTING:
            return current_download_progress + "\n"
        if media_resource.download_status == DownloadStatusCodes.DONE:
            return current_download_progress

        progress_bar = self.format_download_progress_bar(video_index)

        current_download_progress += "\n" + progress_bar + "\n\n"

        return current_download_progress

    def format_download_progress_heading(self, video_index: int):
        media_resource = self.store.downloading_media_resources[video_index]

        status_character = self.get_next_status_character(
            video_index, media_resource.download_status
        )
        status_color = self.get_status_color(media_resource.download_status)
        label = media_resource.download_status.title()

        heading = (
            "\n"
            + colored(f"{status_character} {label} ", fore=status_color)
            + colored(
                f"{self.format_video_title(video_index)} ",
                style=Colors.style.BRIGHT,
            )
            + colored(
                f"{self.store.downloading_media_resources[video_index].formatted_definition}",
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT,
            )
        )

        return heading

    def format_download_progress_bar(self, video_index: int):
        media_resource = self.store.downloading_media_resources[video_index]
        status_color = self.get_status_color(media_resource.download_status)

        bytes_downloaded = (
            media_resource.total_size - media_resource.get_total_bytes_remaining()
        )

        progress_percentage = bytes_downloaded / media_resource.total_size
        available_width = min(screen.get_console_width(), self.store.MAX_SCREEN_WIDTH)

        formatted_bytes_downloaded = self.format_data_size(bytes_downloaded)
        formatted_total_size = self.format_data_size(media_resource.total_size)

        progress_bar_right = colored(
            f" {progress_percentage * 100:.1f}% ",
            fore=status_color,
            style=Colors.style.BRIGHT,
        ) + colored(
            f"({formatted_bytes_downloaded} / " + f"{formatted_total_size})",
            style=Colors.style.DIM,
        )

        max_progress_bar_right_width = len(
            f" 100.0% ({formatted_total_size} / {formatted_total_size})"
        )

        available_space_for_loading_bar = (
            available_width - max_progress_bar_right_width - 2
        )
        loaded_section_length = floor(
            available_space_for_loading_bar * progress_percentage
        )
        not_loaded_section_length = (
            available_space_for_loading_bar - loaded_section_length
        )

        if global_config.ui_colors_disabled:
            loaded_section_character = "#"
            not_loaded_section_character = "-"
            wrapper_characters = {
                "left": "[",
                "right": "]",
            }
        else:
            loaded_section_character = "█"
            not_loaded_section_character = "█"
            wrapper_characters = {
                "left": "",
                "right": "",
            }

        progress_bar_left = (
            "  "
            + wrapper_characters["left"]
            + colored(
                loaded_section_character * loaded_section_length, fore=status_color
            )
            + colored(
                not_loaded_section_character * not_loaded_section_length,
                style=Colors.style.DIM,
            )
            + wrapper_characters["right"]
        )

        progress_bar = progress_bar_left + progress_bar_right

        return progress_bar

    def format_download_formats(self):
        formatted_definitions = map(
            lambda media_resource: media_resource.formatted_definition,
            self.store.media_resources_to_download,
        )
        formatted_download_formats = " ".join(formatted_definitions)
        return formatted_download_formats

    def get_next_status_character(
        self, video_index: int, download_status: DownloadStatusCodes
    ):
        if download_status == DownloadStatusCodes.DOWNLOADING:
            self.store.rotating_line_frames[video_index] = (
                self.store.rotating_line_frames[video_index] + 1
            ) % 4

            return ROTATING_LINES[self.store.rotating_line_frames[video_index]]

        if download_status == DownloadStatusCodes.CONVERTING:
            self.store.loading_dots_frames[video_index] = (
                self.store.loading_dots_frames[video_index] + 1
            ) % 10

            return LOADING_DOTS[self.store.loading_dots_frames[video_index]]

        if download_status == DownloadStatusCodes.DONE:
            return "✔"

        return ""

    def get_status_color(self, download_status: DownloadStatusCodes):
        if download_status == DownloadStatusCodes.DOWNLOADING:
            return Colors.fore.YELLOW
        if download_status == DownloadStatusCodes.CONVERTING:
            return Colors.fore.CYAN
        if download_status == DownloadStatusCodes.DONE:
            return Colors.fore.GREEN

        return ""

    def format_available_formats(self, video_index: int):
        video_streams = self.store.videos[video_index].streams.filter(type="video")
        audio_streams = self.store.videos[video_index].streams.filter(type="audio")

        available_formats = {
            "video": set(),
            "audio": set(),
        }
        for stream in video_streams:
            if stream.resolution is not None:
                available_formats["video"].add(stream.resolution)
        for stream in audio_streams:
            available_formats["audio"].add(stream.abr)

        available_formats_sorted_by_definition = {
            "video": reversed(sorted(available_formats["video"], key=parse_int)),
            "audio": reversed(sorted(available_formats["audio"], key=parse_int)),
        }

        return available_formats_sorted_by_definition

    def format_skipped_formats_warning(self, video_index: int):
        skipped_formats = self.store.skipped_formats[video_index]

        if len(skipped_formats) == 0:
            return ""

        formatted_skipped_formats = [
            f'[{" ".join(skipped_format)}]'
            for skipped_format in self.store.skipped_formats[video_index]
        ]

        skipped_formats_message = (
            (
                "Formats "
                if len(self.store.skipped_formats[video_index]) > 1
                else "Format "
            )
            + colored(
                " ".join(formatted_skipped_formats),
                fore=Colors.fore.MAGENTA,
                style=Colors.style.BRIGHT,
            )
            + (
                " were "
                if len(self.store.skipped_formats[video_index]) > 1
                else " was "
            )
            + "not found. Skipping "
            + ("them" if len(self.store.skipped_formats[video_index]) > 1 else "it")
            + "...\n"
        )

        skipped_formats_warning = Content.format_inner_text(
            skipped_formats_message,
            ContentCategories.WARNING,
        )

        return skipped_formats_warning

    def format_fallback_replacement_summary(self, video_index: int):
        replacement_summaries = []
        for replaced_format in self.store.formats_replaced_by_fallback[video_index]:
            base_format = replaced_format["base"]["format"]
            base_definition = replaced_format["base"]["definition"]
            fallback_format = replaced_format["fallback"]["format"]
            fallback_definition = replaced_format["fallback"]["definition"]

            formatted_base = self.format_replaced_format(base_format, base_definition)
            formatted_fallback = self.format_replaced_format(
                fallback_format, fallback_definition
            )

            replacement_summaries.append(
                "Format "
                + colored(
                    formatted_base, fore=Colors.fore.YELLOW, style=Colors.style.BRIGHT
                )
                + " is not available for this video. Falling back to "
                + colored(
                    formatted_fallback, fore=Colors.fore.BLUE, style=Colors.style.BRIGHT
                )
                + "\n"
            )

        formatted_replacement_summaries = [
            Content.format_inner_text(summary_text, ContentCategories.WARNING)
            for summary_text in replacement_summaries
        ]

        return "\n".join(formatted_replacement_summaries)

    def format_replaced_format(self, current_format, definition):
        return (
            "["
            + (f"{current_format} " if current_format != "videoaudio" else "")
            + f"{definition}]"
        )

    def format_ready_to_download_label(self, video_index: int):
        formatted_title = limit_text_length(self.store.videos[video_index].title, 26)
        formatted_download_formats = self.format_download_formats()

        is_preceded_by_format_warnings = (
            len(self.store.skipped_formats[video_index]) > 0
            or len(self.store.formats_replaced_by_fallback[video_index]) > 0
        )

        prefix = "\n" if is_preceded_by_format_warnings else ""
        colored_formatted_title = colored(
            f"{formatted_title} ",
            fore=Colors.fore.CYAN,
            style=Colors.style.BRIGHT,
        )
        colored_formatted_download_formats = colored(
            f"{formatted_download_formats}\n",
            fore=Colors.fore.BLUE,
            style=Colors.style.BRIGHT,
        )

        return (
            prefix
            + "Ready to download "
            + colored_formatted_title
            + colored_formatted_download_formats
        )

    def format_total_download_size_label(self, video_index: int):
        media_resource = self.store.media_resources_to_download[video_index]
        total_download_size = media_resource.total_size

        formatted_download_size = self.format_data_size(total_download_size)
        total_download_size_label = Content.format_inner_text(
            "Total download size: "
            + colored(f"{formatted_download_size}\n", fore=Colors.fore.YELLOW),
            ContentCategories.INFO,
        )

        return total_download_size_label

    def format_data_size(self, size_in_bytes):
        size_in_megabytes = size_in_bytes / BYTES_PER_MEGABYTES
        return f"{size_in_megabytes:.1f} MB"



from mediakit.cli.screen import Content


class LoadableCLIStore:
    def __init__(self):
        self.LOADING_UI_UPDATE_INTERVAL = 0.4

        self.is_loading: bool = False
        self.loading_label: Content = None

    def reset(self):
        self.is_loading: bool = False
        self.loading_label: Content = None



from mediakit.cli.loadble_cli_store import LoadableCLIStore
from threading import Thread
from time import sleep

from mediakit.cli.screen import screen, ContentCategories


class LoadableCLI:
    def __init__(self, store: LoadableCLIStore):
        self.store = store

    def mark_as_loading(self, is_loading: bool):
        self.store.is_loading = is_loading

        if is_loading:
            self._show_loading_label()
        else:
            self._remove_loading_label()

    def _show_loading_label(self, loading_label: str = "Loading video"):
        self.store.loading_label = screen.append_content(
            f"\n{loading_label}.\n\n", ContentCategories.INFO
        )

        def run_loading_animation():
            current_total_dots_displayed = 0

            while self.store.is_loading:
                current_total_dots_displayed = (current_total_dots_displayed + 1) % 4

                screen.update_content(
                    self.store.loading_label,
                    f"\n{loading_label}" + "." * current_total_dots_displayed + "\n\n",
                )

                sleep(0.4)

        Thread(target=run_loading_animation).start()

    def _remove_loading_label(self):
        if self.store.loading_label is not None:
            screen.remove_content(self.store.loading_label)

        self.store.loading_label = None
        self.store.is_loading = False



from threading import Thread
from pytube import YouTube

from mediakit.cli.screen import Content, ContentCategories, screen
from mediakit.cli.loadble_cli_store import LoadableCLIStore
from mediakit.media.download import DownloadStatusCodes, MediaResource


class DownloadCLIStore(LoadableCLIStore):
    def __init__(self):
        self.MAX_SCREEN_WIDTH = 80
        self.MAX_SHORT_VIDEO_TITLE_LENGTH = 26

        self.UI_UPDATE_INTERVALS = {
            DownloadStatusCodes.DOWNLOADING: 0.15,
            DownloadStatusCodes.CONVERTING: 0.1,
        }
        self.DEFAULT_UI_UPDATE_INTERVAL = 0.2
        self.PROGRESS_UI_UPDATE_INTERVAL = self.DEFAULT_UI_UPDATE_INTERVAL

        self.video_headings: "list[Content]" = []
        self.ready_to_download_labels: "list[Content]" = []
        self.download_confirmation_prompts: "list[Content]" = []

        self.output_path: str = None
        self.filename: str = None

        self.videos: "list[YouTube]" = []

        self.available_formats = []
        self.media_resources_to_download: "list[MediaResource]" = []
        self.skipped_formats = []
        self.formats_replaced_by_fallback = []

        self.downloading_media_resources: "list[MediaResource]" = []
        self.download_progress_uis: "list[Content]" = []
        self.rotating_line_frames: "list[int]" = []
        self.loading_dots_frames: "list[int]" = []

        self.download_ui_update_thread: Thread = None

        self.is_terminated: bool = False

    def prepare_store(self, number_of_videos: int):
        self.videos = [None for _ in range(number_of_videos)]
        self.available_formats = [
            {"video": [], "audio": []} for _ in range(number_of_videos)
        ]
        self.media_resources_to_download = [None for _ in range(number_of_videos)]
        self.skipped_formats = [[] for _ in range(number_of_videos)]
        self.formats_replaced_by_fallback = [[] for _ in range(number_of_videos)]

        self.downloading_media_resources = [None for _ in range(number_of_videos)]

        self.video_headings = [None for _ in range(number_of_videos)]
        self.ready_to_download_labels = [None for _ in range(number_of_videos)]
        self.download_progress_uis = [None for _ in range(number_of_videos)]
        self.download_confirmation_prompts: "list[Content]" = [
            None for _ in range(number_of_videos)
        ]

        self.rotating_line_frames = [0 for _ in range(number_of_videos)]
        self.loading_dots_frames = [0 for _ in range(number_of_videos)]

    def reset(self):
        super().reset()

        self.video_headings: "list[Content]" = []
        self.ready_to_download_labels: "list[Content]" = []
        self.download_confirmation_prompts: "list[Content]" = []

        self.output_path: str = None
        self.filename: str = None

        self.videos: "list[YouTube]" = []

        self.available_formats = []
        self.media_resources_to_download: "list[MediaResource]" = []
        self.skipped_formats = []
        self.formats_replaced_by_fallback = []

        self.downloading_media_resources: "list[MediaResource]" = []
        self.download_progress_uis: "list[Content]" = []
        self.rotating_line_frames: "list[int]" = []
        self.loading_dots_frames: "list[int]" = []



from os import devnull
from shutil import which
import subprocess


def run_command_in_background(command):
    with open(devnull, "w") as DEVNULL:
        subprocess.run(
            command, shell=True, stdout=DEVNULL, stderr=subprocess.STDOUT, check=True
        )


def is_command_available(command):
    return which(command) is not None



name = "Mediakit"
version = "0.6.0"
description = "Download YouTube videos fast, directly from the command line"
author = "Diego Aquino"
author_email = "diegocruzdeaquino@gmail.com"
url = "https://github.com/diego-aquino/mediakit"
download_url = "https://github.com/diego-aquino/mediakit/archive/v0.6.0.tar.gz"
license = "GPL-3.0"
copyright = "Copyright 2021 Diego Aquino"

temporary_filename = ".mediakit_download"



from .utils.ffmpeg import get_ffmpeg_binary


FFMPEG_BINARY = get_ffmpeg_binary()

DOWNLOAD_FORMATS = {"videoaudio", "audio", "videoonly"}

VIDEO_DEFINITIONS = {
    "max": {"next": "2160p"},
    "2160p": {"next": "1440p"},
    "1440p": {"next": "1080p"},
    "1080p": {"next": "720p"},
    "720p": {"next": "480p"},
    "480p": {"next": "360p"},
    "360p": {"next": "240p"},
    "240p": {"next": "144p"},
    "144p": {"next": None},
}

VIDEO_DEFINITIONS_ALIASES = {"4k": "2160p"}

AUDIO_DEFINITIONS = {
    "max": {"next": "160kbps"},
    "160kbps": {"next": "128kbps"},
    "128kbps": {"next": "70kbps"},
    "70kbps": {"next": "50kbps"},
    "50kbps": {"next": None},
}

AVAILABLE_DEFINITIONS = set()
for definition_group in [
    VIDEO_DEFINITIONS,
    VIDEO_DEFINITIONS_ALIASES,
    AUDIO_DEFINITIONS,
]:
    for definition in definition_group:
        AVAILABLE_DEFINITIONS.add(definition)



import colorama

from mediakit.globals import global_config


class Colors:
    fore = colorama.Fore
    back = colorama.Back
    style = colorama.Style


class ColoredText:
    def __init__(self, text, fore, back, style):
        self.text = text
        self.fore = fore
        self.back = back
        self.style = style

    def __str__(self):
        return (
            f"{self.fore}{self.back}{self.style}"
            f"{self.text}"
            f"{Colors.style.RESET_ALL}"
        )


def colored(
    *values,
    fore: str = Colors.fore.RESET,
    back: str = Colors.back.RESET,
    style: str = Colors.style.NORMAL,
    sep=" ",
):
    joined_text = sep.join(map(lambda value: str(value), values))

    if global_config.ui_colors_disabled:
        return joined_text

    return str(ColoredText(joined_text, fore, back, style))



from os import path
import uuid

from pytube.__main__ import YouTube

from mediakit import info
from mediakit.media.convert import (
    merge_video_and_audio,
    convert_media,
    ConversionOptions,
)
from mediakit.utils.files import get_safe_filename, remove_file, move_file
from mediakit.constants import VIDEO_DEFINITIONS_ALIASES


class DownloadStatusCodes:
    READY = "READY"
    DOWNLOADING = "DOWNLOADING"
    CONVERTING = "CONVERTING"
    DONE = "DONE"


class MediaResource:
    def __init__(
        self,
        source,
        output_type,
        output_path=None,
        definition="max",
        filename="",
        filename_suffix="",
    ):
        self.source = source
        self.output_type = output_type
        self.output_path = output_path
        self.definition = definition

        final_definition = (
            VIDEO_DEFINITIONS_ALIASES.get(definition, definition)
            if output_type.startswith("video")
            else definition
        )
        is_alias_definition = definition != final_definition

        default_extension = "mp4" if self.output_type.startswith("video") else "mp3"

        if filename:
            *initial_filename_components, last_filename_component = filename.split(".")

            extension_was_provided = len(initial_filename_components) > 0

            if extension_was_provided:
                filename_without_extension = (
                    ".".join(initial_filename_components) + filename_suffix
                )
                extension = last_filename_component
                self.filename = get_safe_filename(
                    f"{filename_without_extension}.{extension}"
                )
            else:
                filename_without_extension = last_filename_component + filename_suffix
                self.filename = get_safe_filename(
                    f"{filename_without_extension}.{default_extension}"
                )
        else:
            filename_without_extension = source.title + filename_suffix
            self.filename = get_safe_filename(
                f"{filename_without_extension}.{default_extension}"
            )

        if output_type.startswith("video"):
            self.video = self._get_video_stream(final_definition)
            self.total_size = self.video.filesize
            self.video_bytes_remaining = self.video.filesize

            if output_type == "videoaudio":
                if self._is_audio_included():
                    self.has_external_audio = False
                else:
                    self._include_external_audio()
                    self.has_external_audio = True

        elif output_type == "audio":
            self.audio = self._get_audio_stream(final_definition)
            self.total_size = self.audio.filesize
            self.audio_bytes_remaining = self.audio.filesize

        self.formatted_definition = (
            f"[{output_type} " if output_type != "videoaudio" else "["
        )

        if definition == "max":
            self.formatted_definition += (
                f"{self.video.resolution}]"
                if self.output_type.startswith("video")
                else f"{self.audio.abr}]"
            )
        elif is_alias_definition:
            self.formatted_definition += f"{definition}]"
        else:
            self.formatted_definition += f"{final_definition}]"

        self.download_status = DownloadStatusCodes.READY
        self.downloading_stream = None

        unique_resource_id = uuid.uuid4()
        self.temporary_video_filename = (
            f"{info.temporary_filename}-{unique_resource_id}[video].webm"
        )
        self.temporary_audio_filename = (
            f"{info.temporary_filename}-{unique_resource_id}[audio].webm"
        )

    def download(self):
        self.download_status = DownloadStatusCodes.DOWNLOADING

        if self.output_type.startswith("video"):
            self.downloading_stream = self.video

            self.video.download(
                output_path=self.output_path,
                filename=self.temporary_video_filename,
                skip_existing=False,
            )
        if self.output_type == "audio" or (
            self.output_type == "videoaudio" and self.has_external_audio
        ):
            self.downloading_stream = self.audio

            self.audio.download(
                output_path=self.output_path,
                filename=self.temporary_audio_filename,
                skip_existing=False,
            )

        self.download_status = DownloadStatusCodes.CONVERTING
        self._convert_dowloaded_resources()
        self.download_status = DownloadStatusCodes.DONE

    def get_total_bytes_remaining(self):
        if self.output_type == "videoaudio":
            if self.has_external_audio:
                return self.video_bytes_remaining + self.audio_bytes_remaining

            return self.video_bytes_remaining

        if self.output_type == "videoonly":
            return self.video_bytes_remaining

        if self.output_type == "audio":
            return self.audio_bytes_remaining

    def copy(self, source: YouTube = None):
        final_source = source if source is not None else self.source

        return MediaResource(
            final_source,
            self.output_type,
            output_path=self.output_path,
            definition=self.definition,
            filename=self.filename,
        )

    def _convert_dowloaded_resources(self):
        if self.output_type.startswith("video"):
            if self.output_type == "videoaudio" and self.has_external_audio:
                self._merge_video_with_external_audio()
            else:
                self._convert_downloaded_video()
        else:
            self._convert_downloaded_audio()

    def _merge_video_with_external_audio(self):
        video_path = path.join(self.output_path, self.temporary_video_filename)
        audio_path = path.join(self.output_path, self.temporary_audio_filename)
        output_file_path = path.join(self.output_path, self.filename)

        merge_video_and_audio(video_path, audio_path, output_file_path)

        remove_file(video_path)
        remove_file(audio_path)

    def _convert_downloaded_video(self):
        downloaded_temp_file_path = path.join(
            self.output_path, self.temporary_video_filename
        )
        output_file_path = path.join(self.output_path, self.filename)

        options = []
        if self.output_type == "videoonly":
            options.append(ConversionOptions.NO_AUDIO)

        convert_media(
            downloaded_temp_file_path, output_file_path, "mp4", options=options
        )
        remove_file(downloaded_temp_file_path)

    def _convert_downloaded_audio(self):
        downloaded_temp_filename = self.temporary_audio_filename

        downloaded_temp_file_path = path.join(
            self.output_path, downloaded_temp_filename
        )
        output_file_path = path.join(self.output_path, self.filename)

        convert_media(downloaded_temp_file_path, output_file_path, "mp3")
        remove_file(downloaded_temp_file_path)

    def _is_audio_included(self):
        return self.video.is_progressive

    def _get_video_stream(self, definition):
        if definition == "max":
            video_stream = self.source.streams.filter(type="video").order_by(
                "resolution"
            )[-1]
        else:
            video_stream = self.source.streams.filter(
                type="video", resolution=definition
            )[-1]

        return video_stream

    def _get_audio_stream(self, definition):
        if definition == "max":
            audio_stream = self.source.streams.filter(type="audio").order_by("abr")[-1]
        else:
            audio_stream = self.source.streams.filter(type="audio", abr=definition)[-1]

        return audio_stream

    def _include_external_audio(self):
        self.audio = self._get_audio_stream("max")

        self.total_size += self.audio.filesize
        self.audio_bytes_remaining = self.audio.filesize



from mediakit.cli.screen import screen, ContentCategories
from mediakit.cli.colors import colored, Colors


class MediakitException(Exception):
    def __init__(self, message, category):
        self.message = message
        self.category = category

    def show_message(self):
        screen.append_content(self.message, self.category)


class CommandUnavailable(MediakitException):
    def __init__(self, command):
        message = f"Command '{command}' is not available.\n\n"
        category = ContentCategories.ERROR

        super().__init__(message, category)


class FFMPEGNotAvailable(MediakitException):
    def __init__(self):
        message = (
            "FFmpeg is a required package for Mediakit, "
            "but it doesn't appear to be installed.\n"
            "You can install FFmpeg at https://ffmpeg.org/download.html.\n\n"
        )
        category = ContentCategories.WARNING

        super().__init__(message, category)


class InvalidVideoURLError(MediakitException):
    def __init__(self):
        message = "Could not recognize the provided URL. Please, try again.\n\n"
        category = ContentCategories.ERROR

        super().__init__(message, category)


class UnspecifiedError(MediakitException):
    def __init__(self):
        message = "Something went wrong. :(\nPlease, try again.\n\n"
        category = ContentCategories.ERROR

        super().__init__(message, category)


class NoAvailableSpecifiedFormats(MediakitException):
    def __init__(self, available_formats):
        message = (
            "None of the specified formats were found.\n"
            + "Please, verify your entries and try again.\n\n"
            + "The formats available for this video are:\n"
            + colored("   (video) ", style=Colors.style.DIM + Colors.style.BRIGHT)
            + colored(
                " ".join(available_formats["video"]),
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT,
            )
            + colored("\n   (audio) ", style=Colors.style.DIM + Colors.style.BRIGHT)
            + colored(
                " ".join(available_formats["audio"]),
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT,
            )
            + "\n\n"
        )
        category = ContentCategories.ERROR

        super().__init__(message, category)


class NoSuchFile(MediakitException):
    def __init__(self, supposed_file):
        message = (
            "\nNo such file: "
            + colored(supposed_file, fore=Colors.fore.MAGENTA)
            + "\n\n"
        )
        category = ContentCategories.ERROR

        super().__init__(message, category)


class NoVideoURLsInBatchFile(MediakitException):
    def __init__(self, batch_file):
        message = (
            "\nCould not find any valid video URLs in "
            + colored(batch_file, fore=Colors.fore.MAGENTA)
            + "\nPlease check your entries and try again.\n\n"
        )
        category = ContentCategories.ERROR

        super().__init__(message, category)



from mediakit.actions import download
from mediakit.cli.arguments import (
    command_args,
    GlobalArguments,
    show_help_message,
    show_current_version,
    show_arguments_not_recognized_error,
)


def get_command_actions():
    actions = []

    if command_args.has_argument(GlobalArguments.help):
        actions.append(show_help_message)
    if command_args.has_argument(GlobalArguments.version):
        actions.append(show_current_version)
    if command_args.has_video_url() or command_args.has_argument(GlobalArguments.batch):
        actions.append(download)

    if len(actions) == 0:
        if len(command_args.arguments) > 0:
            actions.append(show_arguments_not_recognized_error)

        actions.append(show_help_message)

    return actions



import re


ANSI_ESCAPE_CODES_REGEX = r"\x1b\[[;\d]*[A-Za-z]"

VIDEO_ID_REGEX = r"(?:v=|\/)([0-9A-Za-z_-]{11}).*"
YOUTUBE_URL_REGEX = r"https:\/\/.*youtu"


def search(pattern, string, group):
    regex = re.compile(pattern)
    results = regex.search(string)

    if not results:
        return None

    return results.group(group)


def match(pattern, string, group):
    regex = re.compile(pattern)
    results = regex.match(string)

    if not results:
        return None

    return results.group(group)


def sub(pattern, replacement, string):
    regex = re.compile(pattern)
    replaced_string = regex.sub(replacement, string)

    return replaced_string


def extract_video_id(video_url):
    return search(VIDEO_ID_REGEX, video_url, 1)


def matches_a_youtube_url(string):
    return match(YOUTUBE_URL_REGEX, string, 0)


def is_youtube_video_url(string):
    video_id = extract_video_id(string)
    if video_id is not None and matches_a_youtube_url(string):
        return True

    return False



def flatten_list(list_to_flat: list, depth: int = 1):
    partially_flatten_list = [item for item in list_to_flat]

    for _ in range(depth):
        new_partially_flatten_list = []
        for items in partially_flatten_list:
            if type(items) is list:
                new_partially_flatten_list.extend(items)
            else:
                new_partially_flatten_list.append(items)

        partially_flatten_list = new_partially_flatten_list

    flatten_list = partially_flatten_list
    return flatten_list



from . import regex


def limit_text_length(text, limit_of_caracters):
    if len(text) > limit_of_caracters:
        partial_text_with_ellipsis = f"{text[:limit_of_caracters - 3]}..."

        return partial_text_with_ellipsis

    return text


def len_ansi_safe(string):
    string_striped_of_ansi_codes = regex.sub(regex.ANSI_ESCAPE_CODES_REGEX, "", string)

    return len(string_striped_of_ansi_codes)


def parse_int(string):
    integer_section_as_list = []
    for character in string:
        if character.isnumeric():
            integer_section_as_list.append(character)
        else:
            break

    integer_section = "".join(integer_section_as_list)

    return int(integer_section)



from os import path, listdir
import re

from mediakit.utils.commands import run_command_in_background
from mediakit.info import temporary_filename
from mediakit.utils import regex


def file_exists(file_path):
    return path.isfile(file_path)


def move_file(file_path, new_file_path):
    move_command = f'mv "{file_path}" "{new_file_path}"'

    run_command_in_background(move_command)


def remove_file(file_path):
    remove_command = f'rm -rf "{file_path}"'

    run_command_in_background(remove_command)


def get_safe_filename(filename):
    partially_safe_filename = re.sub(r'[{}\\"\']', "", filename)
    safe_filename = re.sub(r"[/]", " ", partially_safe_filename)

    return safe_filename


def get_filename_from(file_path):
    filename_candidate = path.basename(file_path)
    valid_filename = 0 < filename_candidate.rfind(".") < len(filename_candidate) - 1

    return filename_candidate if valid_filename else ""


def increment_filename_if_exists(file_path):
    file_path_dirname = path.dirname(file_path)

    filename = get_filename_from(file_path)
    extension_start_index = filename.rfind(".")

    filename_without_extension = filename[:extension_start_index]
    extension = filename[extension_start_index + 1 :]

    final_file_path = file_path
    filename_number = 0
    while True:
        if not file_exists(final_file_path):
            return final_file_path

        filename_number += 1
        final_file_path = path.join(
            file_path_dirname,
            f"{filename_without_extension}({filename_number}).{extension}",
        )


def read_video_urls_from(batch_file_path):
    video_urls = []

    with open(batch_file_path, "r", encoding="utf-8") as batch_file:
        for line in batch_file:
            clear_line = line.strip()

            if regex.is_youtube_video_url(clear_line):
                video_urls.append(clear_line)

    return video_urls


def get_temporary_files(directory_path: str) -> "list[str]":
    items_in_directory = [
        path.join(directory_path, item) for item in listdir(directory_path)
    ]

    files_in_directory = filter(path.isfile, items_in_directory)

    temporary_files = filter(
        lambda file_path: get_filename_from(file_path).startswith(temporary_filename),
        files_in_directory,
    )

    return list(temporary_files)


def remove_all_temporary_files(directory_path: str) -> None:
    for temporary_filename in get_temporary_files(directory_path):
        remove_file(temporary_filename)



from os import path

from setuptools import setup, find_packages

from mediakit import info

current_dirname = path.dirname(path.abspath(__file__))
readme_path = path.join(current_dirname, "README.md")

with open(readme_path, encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name=info.name.lower(),
    version=info.version,
    description=info.description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=info.author,
    author_email=info.author_email,
    url=info.url,
    download_url=info.download_url,
    license=info.license,
    packages=find_packages(),
    include_package_data=True,
    keywords=[
        "youtube",
        "media",
        "video",
        "audio",
        "download",
        "convert",
        "command line",
        "cli",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pytube>=11.0.1",
        "clint",
        "imageio",
        "imageio-ffmpeg",
        "colorama",
        "wheel",
    ],
    entry_points={"console_scripts": ["mediakit=mediakit:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Multimedia :: Video",
        "Topic :: Multimedia :: Video :: Conversion",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
    ],
)



import sys


class GlobalConfig:
    def __init__(self):
        self.answer_yes_to_all_questions = False
        self.ui_colors_disabled = not sys.stdin.isatty()
        self.batch_file = None
        self.max_downloads_in_parallel = 2


global_config = GlobalConfig()



import sys
from math import ceil

from clint.textui.cols import console_width

from mediakit.utils.format import len_ansi_safe
from .colors import colored, Colors


class ANSIConsoleExpressions:
    MOVE_CURSOR_ONE_LINE_UP = "\033[A"
    MOVE_CURSOR_TO_BEGGINING_OF_LINE = "\033[0G"
    CLEAR_CHARACTERS_TO_THE_RIGHT_OF_CURSOR = "\033[K"
    CLEAR_LINE = (
        MOVE_CURSOR_TO_BEGGINING_OF_LINE + CLEAR_CHARACTERS_TO_THE_RIGHT_OF_CURSOR
    )


class ContentCategories:
    NORMAL = "NORMAL"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    USER_INPUT = "USER_INPUT"


class Content:
    def __init__(self, index_on_screen, text, category):
        self.index_on_screen = index_on_screen
        self.update(text, category)

    def update(self, text: str = None, category: str = None):
        if category is not None:
            self._update_category(category)
        if text is not None:
            self._update_inner_text(text)

    def is_empty(self):
        return self.inner_text == ""

    def _update_category(self, category: str):
        self.category = category

    def _update_inner_text(self, text):
        self.inner_text = Content.format_inner_text(text, self.category)

    @staticmethod
    def format_category_label(category: str):
        if category == ContentCategories.NORMAL:
            return ""
        elif category == ContentCategories.INFO:
            return colored("info ", fore=Colors.fore.BLUE)
        elif category == ContentCategories.WARNING:
            return colored("warning ", fore=Colors.fore.YELLOW)
        elif category == ContentCategories.ERROR:
            return colored("error ", fore=Colors.fore.RED)
        elif category == ContentCategories.USER_INPUT:
            return colored("? ", fore=Colors.fore.BLUE)

        return ""

    @staticmethod
    def format_inner_text(text: str, category: str):
        if len(text) == 0:
            return ""

        new_text_stripped_of_leading_newlines = text.lstrip("\n")
        number_of_leading_newlines = len(text) - len(
            new_text_stripped_of_leading_newlines
        )

        return (
            "\n" * number_of_leading_newlines
            + Content.format_category_label(category)
            + new_text_stripped_of_leading_newlines
        )


class Screen:
    def __init__(self):
        self.contents = []

    def append_content(self, content_text, category=ContentCategories.NORMAL):
        index_on_screen = len(self.contents)

        new_content = Content(index_on_screen, content_text, category)

        self.contents.append(new_content)
        self._render_content(new_content)

        return new_content

    def update_content(
        self,
        content: Content,
        new_content_text: str,
        new_category: ContentCategories = None,
    ):
        self._clear_lines_starting_at(content)
        content.update(new_content_text, new_category)
        self._render_contents_starting_at(content)

    def remove_content(self, content):
        index_to_remove_at = content.index_on_screen

        self._clear_lines_starting_at(content)
        self.contents.pop(index_to_remove_at)

        for index in range(index_to_remove_at, len(self.contents)):
            self.contents[index].index_on_screen = index

        needs_to_rerender_contents = len(
            self.contents
        ) > 0 and index_to_remove_at < len(self.contents)

        if needs_to_rerender_contents:
            self._render_contents_starting_at(self.contents[index_to_remove_at])

    def clear_lines(self, number_of_lines_to_clear):
        clear_expression = ANSIConsoleExpressions.MOVE_CURSOR_ONE_LINE_UP.join(
            [ANSIConsoleExpressions.CLEAR_LINE] * number_of_lines_to_clear
        )

        print(clear_expression, end="", file=sys.stdout)

    def get_console_width(self):
        return console_width({})

    def prompt(
        self,
        message,
        valid_inputs=[],
        case_sensitive=False,
        index_on_screen=None,
    ):
        if not case_sensitive:
            for index in range(len(valid_inputs)):
                valid_inputs[index] = valid_inputs[index].lower()

        valid_inputs = set(valid_inputs)

        if index_on_screen is None:
            prompt_message = self.append_content(
                message, category=ContentCategories.USER_INPUT
            )
        else:
            prompt_message = self.contents[index_on_screen]
            self.update_content(
                prompt_message, message, new_category=ContentCategories.USER_INPUT
            )

        while True:
            entry = input().strip()

            if not case_sensitive:
                entry = entry.lower()

            valid_entry = len(valid_inputs) == 0 or entry in valid_inputs

            if valid_entry:
                return entry

            self.erase_prompt_entry(prompt_message)

    def erase_prompt_entry(self, prompt):
        lines_occupied_by_entry = 2  # <entry>\n<empty line> -> 2 lines to clear
        self.clear_lines(lines_occupied_by_entry)

        self._clear_lines_starting_at(prompt)
        self._render_contents_starting_at(prompt)

    def _render_content(self, content):
        print(content.inner_text, end="", file=sys.stdout)

    def _clear_lines_starting_at(self, content):
        current_console_width = self.get_console_width()

        texts_to_clear = map(
            lambda content_to_clear: content_to_clear.inner_text,
            self.contents[content.index_on_screen :],
        )

        lines_to_clear = self._count_lines_occupied_by(
            "".join(texts_to_clear), current_console_width
        )

        self.clear_lines(lines_to_clear)

    def _render_contents_starting_at(self, content):
        for index in range(content.index_on_screen, len(self.contents)):
            self._render_content(self.contents[index])

    def _count_lines_occupied_by(self, text, current_console_width):
        lines = text.split("\n")

        lines_occupied = 0
        for i in range(len(lines)):
            line = lines[i]

            lines_occupied += max(ceil(len_ansi_safe(line) / current_console_width), 1)

        return lines_occupied


screen = Screen()



from os import path

from pytube.exceptions import RegexMatchError as PytubeRegexMatchError

from mediakit.cli.arguments import command_args
from mediakit.cli.download import DownloadCLI
from mediakit.utils.files import (
    get_filename_from,
    read_video_urls_from,
    file_exists,
    remove_all_temporary_files,
)
from mediakit.globals import global_config
from mediakit import exceptions


def _get_video_urls_to_download(arguments):
    was_batch_file_provided = global_config.batch_file is not None

    if not was_batch_file_provided:
        return [arguments.video_url]

    if not file_exists(global_config.batch_file):
        raise exceptions.NoSuchFile(global_config.batch_file)

    video_urls_to_download = read_video_urls_from(global_config.batch_file)

    if len(video_urls_to_download) == 0:
        raise exceptions.NoVideoURLsInBatchFile(global_config.batch_file)

    return video_urls_to_download


def download():
    arguments = command_args.parse_download_arguments()

    try:
        video_urls_to_download = _get_video_urls_to_download(arguments)
    except (exceptions.NoSuchFile, exceptions.NoVideoURLsInBatchFile) as exception:
        exception.show_message()
        return

    output_path = path.abspath(arguments.output_path)
    filename = get_filename_from(output_path)
    if filename:
        output_path = path.dirname(output_path)

    download_cli = DownloadCLI(output_path, filename)

    try:
        download_cli.start(video_urls_to_download, arguments.formats)
        download_cli.download_all()
    except exceptions.NoAvailableSpecifiedFormats as exception:
        exception.show_message()
    except PytubeRegexMatchError:
        download_cli.mark_as_loading(False)
        exceptions.InvalidVideoURLError().show_message()
    except KeyboardInterrupt:
        download_cli.terminate()
    except Exception as exception:
        download_cli.mark_as_loading(False)
        exceptions.UnspecifiedError().show_message()
        raise exception
    finally:
        remove_all_temporary_files(output_path)
        download_cli.terminate()



import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from mediakit.info import name, version, description
from mediakit.utils import regex
from mediakit.cli.colors import colored, Colors
from mediakit.cli.screen import screen, ContentCategories
from mediakit.globals import global_config


class GlobalArguments:
    help = ["-h", "--help"]
    yes = ["-y", "--yes"]
    version = ["-v", "--version"]
    no_colors = ["-nc", "--no-colors"]
    batch = ["-b", "--batch"]
    max_downloads_in_parallel = ["-p", "--parallel"]


class Parser(ArgumentParser):
    def add_global_arguments(self):
        self.add_argument(
            *GlobalArguments.help, action="store_true", help="Show this help message"
        )
        self.add_argument(
            *GlobalArguments.yes,
            action="store_true",
            help='Answer "yes" to all questions beforehand',
        )
        self.add_argument(
            *GlobalArguments.version,
            action="store_true",
            help="Show the current version",
        )
        self.add_argument(
            *GlobalArguments.no_colors,
            dest="no_colors",
            action="store_true",
            help="Disable the colors of the interface",
        )
        self.add_argument(
            *GlobalArguments.batch,
            nargs=1,
            help="Download videos from URLs stored in a batch file",
        )
        self.add_argument(
            *GlobalArguments.max_downloads_in_parallel,
            dest="max_downloads_in_parallel",
            type=int,
            default=global_config.max_downloads_in_parallel,
            help="Set a limit to downloads in parallel",
        )

    def add_download_arguments(self):
        if not global_config.batch_file:
            self.add_argument("video_url", help="URL of the YouTube video to download")
        self.add_argument(
            "output_path",
            nargs="?",
            default="./",
            help="Destination folder to where to save the downloads",
        )
        self.add_argument(
            "-f",
            "--formats",
            nargs="*",
            default=[],
            help="Formats to download, separated by spaces (e.g. 1080p 720p 360p)",
        )


def update_global_config_based_on_arguments(arguments):
    global_config.answer_yes_to_all_questions = getattr(
        arguments, "yes", global_config.answer_yes_to_all_questions
    )

    global_config.ui_colors_disabled = getattr(
        arguments, "no_colors", global_config.ui_colors_disabled
    )

    batch_file = getattr(arguments, "batch", global_config.batch_file)
    was_batch_file_provided = isinstance(batch_file, list) and len(batch_file) == 1
    global_config.batch_file = (
        batch_file[0] if was_batch_file_provided else global_config.batch_file
    )

    global_config.max_downloads_in_parallel = arguments.max_downloads_in_parallel


class CommandArgs:
    def __init__(self, args=sys.argv, parse_and_update_global_config=True):
        self.arguments = args[1:]
        self.unique_arguments = set(self.arguments)

        self.parser = Parser(
            formatter_class=RawDescriptionHelpFormatter, add_help=False
        )

        self.parser.add_global_arguments()
        self.global_arguments = self.parser.parse_known_args()[0]
        if parse_and_update_global_config:
            update_global_config_based_on_arguments(self.global_arguments)

        self.parser.description = (
            colored(f"{name.lower()} v{version}\n", style=Colors.style.BRIGHT)
            + description
        )

    def has_argument(self, possible_arguments):
        intersection = self.unique_arguments.intersection(set(possible_arguments))

        return len(intersection) > 0

    def has_video_url(self):
        for argument in self.unique_arguments:
            if regex.is_youtube_video_url(argument):
                return True

        return False

    def parse_download_arguments(self):
        self.parser.add_download_arguments()
        arguments = self.parser.parse_args()
        return arguments


def show_current_version():
    screen.append_content(f"{version}\n")


def show_help_message():
    command_args.parser.add_download_arguments()
    screen.append_content(command_args.parser.format_help())


def show_arguments_not_recognized_error():
    screen.append_content(
        "\nCould not recognize the provided arguments. "
        + "Please check your entries and try again.\n\n",
        ContentCategories.ERROR,
    )


command_args = CommandArgs()



from os import path
from functools import partial
from threading import Thread
from time import sleep
from typing import Callable
from pytube import YouTube

from mediakit.cli.screen import screen, ContentCategories
from mediakit.cli.colors import colored, Colors
from mediakit.media.download import DownloadStatusCodes, MediaResource
from mediakit.utils.format import parse_int
from mediakit.cli.loadble_cli import LoadableCLI
from mediakit.cli.download.download_cli_formatter import DownloadCLIFormatter
from mediakit.cli.download.download_cli_store import DownloadCLIStore
from mediakit.constants import (
    DOWNLOAD_FORMATS,
    FFMPEG_BINARY,
    VIDEO_DEFINITIONS,
    VIDEO_DEFINITIONS_ALIASES,
    AUDIO_DEFINITIONS,
    AVAILABLE_DEFINITIONS,
)
from mediakit.globals import global_config
from mediakit import exceptions
from mediakit.utils.lists import flatten_list


class DownloadCLI(LoadableCLI):
    def __init__(self, output_path: str, filename: str):
        self.store = DownloadCLIStore()
        self.store.filename = filename
        self.store.output_path = output_path

        self.formatter = DownloadCLIFormatter(self.store)

    def start(self, video_urls_to_download: "list[str]", formats):
        self._show_header(video_urls_to_download)

        if not FFMPEG_BINARY:
            exceptions.FFMPEGNotAvailable().show_message()
            return

        self.mark_as_loading(True, len(video_urls_to_download))
        self._register_videos_to_download(video_urls_to_download, formats)
        self.mark_as_loading(False)

    def mark_as_loading(self, is_loading: bool, number_of_videos_being_loaded=1):
        self.store.is_loading = is_loading

        if is_loading:
            loading_label = (
                "Loading videos"
                if number_of_videos_being_loaded > 1
                else "Loading video"
            )
            self._show_loading_label(loading_label)
        else:
            self._remove_loading_label()

    def _show_header(self, video_urls_to_download):
        screen.append_content(self.formatter.format_header())

        if global_config.batch_file:
            self._show_batch_file_info(video_urls_to_download)

    def _register_videos_to_download(self, video_urls: str, formats: "list[str]"):
        videos = [YouTube(video_url) for video_url in video_urls]
        videos_data = self._load_videos_data(videos, formats)

        available_formats_list: list = []
        media_resources_list: "list[list[MediaResource]]" = []
        skipped_formats_list: list = []
        formats_replaced_by_fallback_list: list = []

        for (
            available_formats,
            media_resources_to_download,
            skipped_formats,
            formats_replaced_by_fallback,
        ) in videos_data:
            available_formats_list.append(available_formats)
            media_resources_list.append(media_resources_to_download)
            skipped_formats_list.append(skipped_formats)
            formats_replaced_by_fallback_list.append(formats_replaced_by_fallback)

        non_flatten_videos = [[video] for video in videos]
        non_flatten_available_formats = [
            [available_format] for available_format in available_formats_list
        ]
        non_flatten_media_resources = [
            media_resources for media_resources in media_resources_list
        ]
        non_flatten_skipped_formats = [
            [skipped_formats] for skipped_formats in skipped_formats_list
        ]
        non_flatten_formats_replaced_by_fallback = [
            [replaced_formats] for replaced_formats in formats_replaced_by_fallback_list
        ]

        for video_index in range(len(non_flatten_videos)):
            source_video = non_flatten_videos[video_index][0]
            available_formats = non_flatten_available_formats[video_index][0]
            media_resources_to_download = non_flatten_media_resources[video_index]
            skipped_formats = non_flatten_skipped_formats[video_index][0]
            replaced_formats = non_flatten_formats_replaced_by_fallback[video_index][0]

            non_flatten_videos[video_index] = [
                YouTube(source_video.watch_url) for _ in media_resources_to_download
            ]
            non_flatten_media_resources[video_index] = [
                media_resources_to_download[media_resource_index].copy(
                    source=non_flatten_videos[video_index][media_resource_index]
                )
                for media_resource_index in range(len(media_resources_to_download))
            ]
            non_flatten_available_formats[video_index] = [
                available_formats for _ in media_resources_to_download
            ]
            non_flatten_skipped_formats[video_index] = [
                skipped_formats for _ in media_resources_to_download
            ]
            non_flatten_formats_replaced_by_fallback[video_index] = [
                replaced_formats for _ in media_resources_to_download
            ]

        flatten_videos = flatten_list(non_flatten_videos)
        self.store.prepare_store(len(flatten_videos))
        self.store.videos = flatten_videos

        self.store.media_resources_to_download = flatten_list(
            non_flatten_media_resources
        )
        self.store.available_formats = flatten_list(non_flatten_available_formats)
        self.store.skipped_formats = flatten_list(non_flatten_skipped_formats)
        self.store.formats_replaced_by_fallback = flatten_list(
            non_flatten_formats_replaced_by_fallback
        )

        for video_index in range(len(self.store.videos)):
            video = self.store.videos[video_index]

            def handle_download_progress(video_index, _stream, _chunk, bytes_remaining):
                self._handle_download_progress(video_index, bytes_remaining)

            video.register_on_progress_callback(
                partial(handle_download_progress, video_index)
            )

    def download_all(self):
        self._prerender_download_ui_components()

        at_least_one_file_was_downloaded = False

        self.store.download_ui_update_thread = Thread(
            target=self._keep_download_progress_ui_updated
        )
        self.store.download_ui_update_thread.start()

        video_indexes_left_to_download = list(range(len(self.store.videos) - 1, -1, -1))
        ongoing_download_threads: "list[Thread]" = []

        def download(video_index: int, on_finish: Callable[..., None]):
            nonlocal at_least_one_file_was_downloaded

            media_resource = self.store.media_resources_to_download[video_index]

            self._create_download_progress(video_index, media_resource)

            media_resource.download()

            self._end_download_progress(video_index)
            at_least_one_file_was_downloaded = True

            on_finish()

        def start_next_download():
            if len(video_indexes_left_to_download) == 0:
                return

            video_index = video_indexes_left_to_download.pop()

            self._show_download_summary(video_index)

            should_ask_confirmation_to_download = (
                not global_config.answer_yes_to_all_questions
                and len(self.store.videos) == 1
            )
            if should_ask_confirmation_to_download:
                user_has_confirmed = self._ask_for_confirmation_to_download(video_index)
                if not user_has_confirmed:
                    return

            download_thread = Thread(
                target=partial(download, video_index, start_next_download)
            )
            ongoing_download_threads.append(download_thread)
            download_thread.start()

        number_of_starter_videos = min(
            global_config.max_downloads_in_parallel,
            len(video_indexes_left_to_download),
        )

        for _ in range(number_of_starter_videos):
            start_next_download()

        for thread in ongoing_download_threads:
            thread.join()

        if at_least_one_file_was_downloaded:
            self._show_success_message()

    def _show_download_summary(self, video_index: int):
        screen.update_content(
            self.store.video_headings[video_index],
            self.formatter.format_download_summary(video_index),
        )

        if self._were_all_formats_skipped(video_index):
            raise exceptions.NoAvailableSpecifiedFormats(
                self.store.available_formats[video_index]
            )

        if len(self.store.videos) == 1:
            screen.update_content(
                self.store.ready_to_download_labels[video_index],
                self.formatter.format_ready_to_download_label(video_index),
            )

    def _ask_for_confirmation_to_download(self, video_index: int):
        prompt = self.store.download_confirmation_prompts[video_index]

        prompt_message = self.formatter.format_download_confirmation_prompt_message()
        answer = screen.prompt(
            prompt_message,
            valid_inputs=["", "y", "n"],
            index_on_screen=prompt.index_on_screen,
        )

        return answer != "n"

    def _create_download_progress(
        self, video_index: int, media_resource: MediaResource
    ):
        self.store.downloading_media_resources[video_index] = media_resource
        screen.update_content(
            self.store.download_progress_uis[video_index],
            self.formatter.format_current_download_progress(video_index),
        )

    def _prerender_download_ui_components(self):
        for video_index in range(len(self.store.videos)):
            self.store.video_headings[video_index] = screen.append_content("")
            self.store.ready_to_download_labels[video_index] = screen.append_content("")
            self.store.download_confirmation_prompts[
                video_index
            ] = screen.append_content("")
            self.store.download_progress_uis[video_index] = screen.append_content("")

    def _keep_download_progress_ui_updated(self):
        has_detailed_download_info_on_screen = [True for _ in self.store.videos]

        while not self.store.is_terminated:
            self._update_all_download_progress_uis(has_detailed_download_info_on_screen)
            sleep(self.store.PROGRESS_UI_UPDATE_INTERVAL)

    def _update_all_download_progress_uis(
        self, has_detailed_download_info_on_screen: "list[bool]"
    ):
        for video_index in range(len(self.store.videos)):
            media_resource = self.store.media_resources_to_download[video_index]

            should_clear_detailed_download_info = (
                media_resource.download_status == DownloadStatusCodes.DONE
                and has_detailed_download_info_on_screen[video_index]
            )

            if should_clear_detailed_download_info:
                self._clear_detailed_download_info_from_screen(video_index)
                has_detailed_download_info_on_screen[video_index] = False

        videos_with_downloading_media_resources = [
            video_index
            for video_index in range(len(self.store.videos))
            if self.store.downloading_media_resources[video_index] is not None
        ]

        for video_index in videos_with_downloading_media_resources:
            self._update_download_progress_ui(video_index)
            self._update_progress_ui_interval_if_necessary(video_index)

    def _update_download_progress_ui(self, video_index: int):
        screen.update_content(
            self.store.download_progress_uis[video_index],
            self.formatter.format_current_download_progress(video_index),
        )

    def _update_progress_ui_interval_if_necessary(self, video_index: int):
        media_resource = self.store.downloading_media_resources[video_index]
        required_interval = self.store.UI_UPDATE_INTERVALS.get(
            media_resource.download_status, self.store.DEFAULT_UI_UPDATE_INTERVAL
        )

        if self.store.PROGRESS_UI_UPDATE_INTERVAL != required_interval:
            self.store.PROGRESS_UI_UPDATE_INTERVAL = required_interval

    def _handle_download_progress(self, video_index: int, bytes_remaining: int):
        media_resource = self.store.downloading_media_resources[video_index]

        if media_resource is None:
            return

        downloading_stream = media_resource.downloading_stream

        if media_resource.output_type == "videoaudio":
            if downloading_stream is media_resource.video:
                media_resource.video_bytes_remaining = bytes_remaining
            elif downloading_stream is media_resource.audio:
                media_resource.audio_bytes_remaining = bytes_remaining

        if media_resource.output_type == "videoonly":
            media_resource.video_bytes_remaining = bytes_remaining

        if media_resource.output_type == "audio":
            media_resource.audio_bytes_remaining = bytes_remaining

    def _end_download_progress(self, video_index: int):
        self._update_download_progress_ui(video_index)

        self.store.downloading_media_resources[video_index] = None
        self.store.download_progress_uis[video_index] = None

    def _show_success_message(self):
        screen.append_content(
            colored("\nSuccess! ", fore=Colors.fore.GREEN)
            + "Files saved at "
            + colored(
                f"{self.store.output_path}/\n\n",
                fore=Colors.fore.CYAN,
            )
        )

    def _clear_detailed_download_info_from_screen(self, video_index: int):
        video_heading = self.store.video_headings[video_index]
        ready_to_download_label = self.store.ready_to_download_labels[video_index]
        confirmation_prompt = self.store.download_confirmation_prompts[video_index]

        screen.update_content(video_heading, "")
        screen.update_content(ready_to_download_label, "")

        if not confirmation_prompt.is_empty():
            screen.erase_prompt_entry(confirmation_prompt)
            screen.update_content(confirmation_prompt, "")

    def terminate(self):
        self._update_all_download_progress_uis([True for _ in self.store.videos])

        self.store.loading_label = None
        self.store.is_loading = False

        self.store.is_terminated = True

        if self.store.download_ui_update_thread is not None:
            self.store.download_ui_update_thread.join()

    def _were_all_formats_skipped(self, video_index: int):
        return (
            self.store.media_resources_to_download[video_index] is None
            and len(self.store.skipped_formats[video_index]) > 0
        )

    def _show_batch_file_info(self, valid_video_urls_in_batch_file):
        batch_filename = path.basename(global_config.batch_file)

        screen.append_content(
            self.formatter.format_batch_file_info_header(batch_filename),
            ContentCategories.INFO,
        )
        screen.append_content(
            self.formatter.format_batch_file_info_body(valid_video_urls_in_batch_file),
            ContentCategories.INFO,
        )

    def _load_videos_data(self, videos: "list[YouTube]", formats):
        videos_data = [None for _ in videos]

        def load_video(video_index: int):
            videos_data[video_index] = self._load_video_data(
                videos[video_index], formats
            )

        loading_threads: "list[Thread]" = []
        for video_index in range(len(videos)):
            thread = Thread(target=partial(load_video, video_index))
            loading_threads.append(thread)
            thread.start()

        for ongoing_thread in loading_threads:
            ongoing_thread.join()

        return videos_data

    def _load_video_data(self, video: YouTube, formats):
        available_formats = self._get_available_formats(video)
        (
            media_resources_to_download,
            skipped_formats,
            formats_replaced_by_fallback,
        ) = self._get_media_resources_to_download(video, formats)

        return (
            available_formats,
            media_resources_to_download,
            skipped_formats,
            formats_replaced_by_fallback,
        )

    def _get_available_formats(self, video: YouTube):
        video_streams = video.streams.filter(type="video")
        audio_streams = video.streams.filter(type="audio")

        available_formats = {
            "video": set(),
            "audio": set(),
        }
        for stream in video_streams:
            if stream.resolution is not None:
                available_formats["video"].add(stream.resolution)
        for stream in audio_streams:
            available_formats["audio"].add(stream.abr)

        available_formats_sorted_by_definition = {
            "video": list(reversed(sorted(available_formats["video"], key=parse_int))),
            "audio": list(reversed(sorted(available_formats["audio"], key=parse_int))),
        }

        return available_formats_sorted_by_definition

    def _get_media_resources_to_download(self, video: YouTube, formats):
        if len(formats) == 0:
            default_media_resource = MediaResource(
                video,
                "videoaudio",
                output_path=self.store.output_path,
                filename=self.store.filename,
            )

            return [default_media_resource], [], []

        lowecased_formats = [selected_format.lower() for selected_format in formats]
        grouped_formats, invalid_formats = self._group_and_validate_formats(
            lowecased_formats
        )

        media_resources_to_download = []
        resources_registered_to_download = set()
        skipped_formats = [invalid_format for invalid_format in invalid_formats]
        formats_replaced_by_fallback = []

        definitions_count = {}
        for group in grouped_formats:
            download_format = group["format"]
            current_count = definitions_count.get(download_format, 0)
            definitions_count[download_format] = current_count + 1

        for group in grouped_formats:
            download_format = group["format"]
            definition = group["definition"]

            should_append_format_to_filename = definitions_count[download_format] > 1

            available_definition = self._get_available_definition(
                video, download_format, definition
            )

            if available_definition is None:
                skipped_formats.append([download_format, definition])
                continue

            if available_definition != definition:
                formats_replaced_by_fallback.append(
                    {
                        "base": {"format": download_format, "definition": definition},
                        "fallback": {
                            "format": download_format,
                            "definition": available_definition,
                        },
                    }
                )

            if (
                download_format,
                available_definition,
            ) in resources_registered_to_download:
                continue

            if should_append_format_to_filename:
                filename_suffix = (
                    f" [{download_format}] " if download_format != "videoaudio" else ""
                ) + (
                    f" [{available_definition}]"
                    if available_definition != "max"
                    else ""
                )
            else:
                filename_suffix = ""

            media_resource = MediaResource(
                video,
                download_format,
                output_path=self.store.output_path,
                definition=available_definition,
                filename=self.store.filename,
                filename_suffix=filename_suffix,
            )

            media_resources_to_download.append(media_resource)
            resources_registered_to_download.add(
                (download_format, available_definition)
            )

        return (
            media_resources_to_download,
            skipped_formats,
            formats_replaced_by_fallback,
        )

    def _group_and_validate_formats(self, formats):
        grouped_formats = []
        invalid_formats = []

        format_index = 0
        while format_index < len(formats):
            current_item = formats[format_index]
            next_item = (
                formats[format_index + 1] if format_index < len(formats) - 1 else None
            )

            is_current_item_a_format = current_item in DOWNLOAD_FORMATS
            is_current_item_a_definition = current_item in AVAILABLE_DEFINITIONS
            is_next_item_a_format = next_item in DOWNLOAD_FORMATS
            is_next_item_a_definition = next_item in AVAILABLE_DEFINITIONS

            if is_current_item_a_format:
                if is_next_item_a_definition:
                    grouped_formats.append(
                        {"format": current_item, "definition": next_item}
                    )
                    format_index += 1
                elif next_item is None or is_next_item_a_format:
                    grouped_formats.append(
                        {"format": current_item, "definition": "max"}
                    )
                else:
                    invalid_formats.append([current_item, next_item])
                    format_index += 1

            elif is_current_item_a_definition:
                grouped_formats.append(
                    {"format": "videoaudio", "definition": current_item}
                )

            else:
                invalid_formats.append([current_item])

            format_index += 1

        return grouped_formats, invalid_formats

    def _get_available_definition(self, video: YouTube, base_format, base_definition):
        if base_format.startswith("video"):
            is_valid_definition = (
                base_definition in VIDEO_DEFINITIONS
                or base_definition in VIDEO_DEFINITIONS_ALIASES
            )
        else:
            is_valid_definition = base_definition in AUDIO_DEFINITIONS

        if not is_valid_definition:
            return None

        possible_definition = base_definition
        while possible_definition is not None:
            if self._is_definition_available(video, base_format, possible_definition):
                return possible_definition

            possible_definition = VIDEO_DEFINITIONS_ALIASES.get(
                possible_definition, possible_definition
            )

            possible_definition = (
                VIDEO_DEFINITIONS[possible_definition]["next"]
                if base_format.startswith("video")
                else AUDIO_DEFINITIONS[possible_definition]["next"]
            )

        return None

    def _is_definition_available(
        self, video: YouTube, base_format: str, definition: str
    ):
        if base_format.startswith("video"):
            final_definition = VIDEO_DEFINITIONS_ALIASES.get(definition, definition)

            if final_definition == "max":
                video_streams = video.streams.filter(type="video")
                return len(video_streams) > 0

            video_streams_with_specified_resolution = video.streams.filter(
                type="video", resolution=final_definition
            )
            return len(video_streams_with_specified_resolution) > 0

        else:
            if definition == "max":
                audio_streams = video.streams.filter(type="audio")
                return len(audio_streams) > 0

            audio_streams_with_specified_resolution = video.streams.filter(
                type="audio", abr=definition
            )
            return len(audio_streams_with_specified_resolution) > 0
